/*
 * Copyright (c) 2001-2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#ifndef _EAP8021X_LINKADDRESSES_H
#define _EAP8021X_LINKADDRESSES_H

/*
 * LinkAddresses.h
 * - get the list of link-level addresses 
 */
/*
 * Modification History
 *
 * November 1, 2001		Dieter Siegmund (dieter@apple.com)
 * - created
 */

#include <sys/socket.h>
#include <net/if.h>
#include <net/if_dl.h>

typedef struct LinkAddresses_s LinkAddresses, * LinkAddressesRef;

LinkAddressesRef
LinkAddresses_create();

void
LinkAddresses_free(LinkAddressesRef * list_p);

struct sockaddr_dl * 
LinkAddresses_lookup(LinkAddressesRef list, char * ifname);

int
LinkAddresses_count(LinkAddressesRef list);

struct sockaddr_dl *
LinkAddresses_element(LinkAddressesRef list, int i);

extern void
sockaddr_dl_print(struct sockaddr_dl * dl_p);

#endif _EAP8021X_LINKADDRESSES_H
