
/*
 * Copyright (c) 2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#ifndef _EAP8021X_RADIUSATTRIBUTES_H
#define _EAP8021X_RADIUSATTRIBUTES_H

/*
 * RADIUS.h
 * - definitions for RADIUS attributes
 */

/* 
 * Modification History
 *
 * October 11, 2002	Dieter Siegmund (dieter@apple)
 * - created
 */

#include <stdint.h>

enum {
    kRADIUSAttributeTypeUserName = 1,
    kRADIUSAttributeTypeUserPassword = 2,
    kRADIUSAttributeTypeCHAPPassword = 3,
    kRADIUSAttributeTypeCHAPChallenge = 60,
    kRADIUSAttributeTypeEAPMessage = 79,
};

typedef uint32_t RADIUSAttributeType;

enum {
    kMSRADIUSAttributeTypeMSCHAPResponse = 1,
    kMSRADIUSAttributeTypeMSCHAPError = 2,
    kMSRADIUSAttributeTypeMSCHAPDomain = 10,
    kMSRADIUSAttributeTypeMSCHAPChallenge = 11,
    kMSRADIUSAttributeTypeMSCHAP2Response = 25,
    kMSRADIUSAttributeTypeMSCHAP2Success = 26,
};
typedef uint32_t MSRADIUSAttributeType;

enum {
    kRADIUSVendorIdentifierMicrosoft = 311,
};

typedef uint32_t RADIUSVendorIdentifer;

#endif _EAP8021X_RADIUSATTRIBUTES_H
